EXCLUDED_TABLES = [
    'sysaggregates',
    'sysams',
    'sysattrtypes',
    'sysautolocate',
    'sysblobs',
    'syscasts',
    'syschecks',
    'syscolattribs',
    'syscolauth',
    'syscoldepend',
    'syscolumns',
    'sysconstraints',
    'sysdefaults',
    'sysdepend',
    'sysdirectives',
    'sysdistrib',
    'syserrors',
    'sysextcols',
    'sysextdfiles',
    'sysexternal',
    'sysfragauth',
    'sysfragdist',
    'sysfragments',
    'sysindices',
    'sysinherits',
    'syslangauth',
    'syslogmap',
    'sysobjstate',
    'sysopclasses',
    'sysopclstr',
    'sysprocauth',
    'sysprocbody',
    'sysproccolumns',
    'sysprocedures',
    'sysprocplan',
    'sysreferences',
    'sysroleauth',
    'sysroutinelangs',
    'sysseclabelauth',
    'sysseclabelcomponentelements',
    'sysseclabelcomponents',
    'sysseclabelnames',
    'sysseclabels',
    'syssecpolicies',
    'syssecpolicycomponents',
    'syssecpolicyexemptions',
    'syssequences',
    'syssurrogateauth',
    'syssynonyms',
    'syssyntable',
    'systabamdata',
    'systabauth',
    'systables',
    'systraceclasses',
    'systracemsgs',
    'systrigbody',
    'systriggers',
    'sysusers',
    'sysviews',
    'sysviolations',
    'sysxadatasources',
    'sysxasourcetypes',
    'sysxtddesc',
    'sysxtdtypeauth',
    'sysxtdtypes'
]
